//		Copyright © 1999  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Screen saver class.

#include <cstdlib>
#include <cstring>

#include <Bitmap.h>
#include <GraphicsDefs.h>
#include <image.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <ScreenSaver.h>
#include <TextView.h>
#include <View.h>

#include "FlipSaver.h"
#include "RandomColor.h"

const rgb_color kBlack = { 0, 0, 0, 0 };

const float kMinimalTileSize = 16;
const float kStrokeWidth = 0.25;
const int kDefaultTiles = 12;

// Screen saver interface.
BScreenSaver *
instantiate_screen_saver(BMessage *pMsg, image_id id)
{
	return new FlipSaver(pMsg, id);
}

FlipSaver::FlipSaver(BMessage *pMsg, image_id id)
:	inherited(pMsg, id),
	mColor(kBlack),
	mTileSize(0),
	mpArray(0),
	mpFlip(0),
	mpFlop(0)
{
	srand((long)system_time());
}

FlipSaver::~FlipSaver(void)
{
	delete [] mpArray;
	mpArray = 0;
	delete mpFlip;
	mpFlip = 0;
	delete mpFlop;
	mpFlop = 0;
}

void
FlipSaver::Draw(BView *pView, int32 frame)
{
	if (!pView || !mpArray || !mpFlip || !mpFlop) {
		return;
	}
	
	const BRect bounds = pView->Bounds();
	const int nX = (int)(bounds.Width() / mTileSize);
	const int nY = (int)(bounds.Height() / mTileSize);
	
	pView->SetLowColor(kBlack);
	pView->SetHighColor(mColor);

	if (frame == 0) {
		// Fill entire screen.
		pView->FillRect(bounds, B_SOLID_LOW);

		bool *pArray = mpArray;
		float y = 0;
		for (int i = 0; i < nY; ++i) {
			float x = 0;
			for (int j = 0; j < nX; ++j) {
				pView->DrawBitmapAsync(*pArray++ ? mpFlip : mpFlop, BPoint(x, y));
				x += mTileSize;
			}
			y += mTileSize;
		}

	} else {
		// Flip a single tile.
		const int i = ((double)rand() / (RAND_MAX+1)) * nY;
		const int j = ((double)rand() / (RAND_MAX+1)) * nX;
		bool *pArray = mpArray + nX * i + j;
		*pArray = !*pArray;
		pView->DrawBitmap(*pArray ? mpFlip : mpFlop,
								BPoint(j * mTileSize, i * mTileSize));
	}

	pView->Sync();
}

void
FlipSaver::StartConfig(BView *pConfigView)
{
	const BRect frame = pConfigView->Bounds().InsetBySelf(10.0, 10.0);
	BTextView *pCaption =
		new BTextView(frame, "FlipSaver", frame.OffsetToCopy(B_ORIGIN),
							B_FOLLOW_ALL, B_WILL_DRAW);
	pCaption->SetText("FlipSaver\n"
							"\n"
							"\t© 1999  Jens Kilian <jjk@acm.org>");
	pConfigView->AddChild(pCaption);
	pCaption->SetViewColor(pConfigView->ViewColor());
	pCaption->SetStylable(true);
	pCaption->SetFontAndColor(be_plain_font);
	const char *pText = pCaption->Text();
	pCaption->SetFontAndColor(0, strchr(pText, '\n') - pText, be_bold_font);
	pCaption->MakeEditable(FALSE);
	pCaption->MakeSelectable(FALSE);
}

status_t
FlipSaver::StartSaver(BView *pView, bool preview)
{
	SetTickSize(preview ? 100000 : 10000);	// 0.1/0.01 second

	mColor = RandomColor(0.5, 1.0, 0.5, 1.0);
	
	// Try default number of tiles, reduce until minimal size is guaranteed.
	const BRect bounds = pView->Bounds();
	const float width = bounds.Width();
	const float height = bounds.Height();

	int nTiles = kDefaultTiles;
	while (nTiles > 0) {
		const float hSize = width / nTiles;
		const float vSize = height / nTiles;
		if (hSize > kMinimalTileSize && vSize > kMinimalTileSize) {
			mTileSize = (hSize > vSize) ? vSize : hSize;
			break;
		}
		--nTiles;
	}

	if (nTiles == 0) {
		// View too small.
		return B_ERROR;
	}

	// Initialize bit array.	
	const int nBits = (int)(width / mTileSize) * (int)(height / mTileSize);
	delete [] mpArray;
	mpArray = new bool[nBits];
	for (int i = 0; i < nBits; ++i) {
		mpArray[i] = ((double)rand() / (RAND_MAX+1)) < 0.5;
	}
	
	// Initialize the two bitmaps.
	const BRect tileRect(0, 0, mTileSize, mTileSize);
	BBitmap canvas(tileRect, B_MONOCHROME_1_BIT, true);
	if (canvas.Lock()) {
		BView cView(tileRect, "cView", B_FOLLOW_NONE, B_WILL_DRAW);
		canvas.AddChild(&cView);
		
		FlipFlop(cView, false);
		delete mpFlip;
		mpFlip = new BBitmap(&canvas);
		
		FlipFlop(cView, true);
		delete mpFlop;
		mpFlop = new BBitmap(&canvas);
		
		canvas.RemoveChild(&cView);
		canvas.Unlock();
	}
	
	return B_OK;
}

void
FlipSaver::FlipFlop(BView &view, bool which)
{
	view.FillRect(BRect(0, 0, mTileSize, mTileSize), B_SOLID_LOW);
	view.SetPenSize(mTileSize * kStrokeWidth);

	const float r = mTileSize * 0.5;
	view.StrokeEllipse(BPoint(0, which ? 0 : mTileSize + 1), r, r);
	view.StrokeEllipse(BPoint(mTileSize + 1, which ? mTileSize + 1 : 0), r, r);
	
	view.Sync();
}

